;created as an attempt to replicate results found by Cremers as expressed in Key Exchange in IPsec Revisited.
;(Or rather from input of ikev1-quick_nopfs.cpp)
(herald "IKEv1 Public-Key" (algebra diffie-hellman) (bound 15))


;Macros to reflect what certian parties do not know
(defmacro (zi) (exp Gr i))
(defmacro (zr) (exp Gi r))
;(defmacro (kir) (hash ni nr (zi))) ; May come back to this or some form of bidirectionality
;(defmacro (kri) (hash  nr ni (zr)))

;Macros for the different messages sent back and forth
(defmacro (hash_1i) (hash "prf" (ltk I R) mid lst ni))
(defmacro (hash_1ii) (hash "prf" (ltk R I) mid lst ni))

(defmacro (hash_1r) (hash "prf" (ltk I R) mid lst ni))
(defmacro (hash_1rr) (hash "prf" (ltk R I) mid lst ni))

(defmacro (hash_2i) (hash "prf" (ltk I R) mid ni algo nr))
(defmacro (hash_2ii) (hash "prf" (ltk R I) mid ni algo nr))

(defmacro (hash_2r) (hash "prf" (ltk I R) mid ni algo nr))
(defmacro (hash_2rr) (hash "prf" (ltk R I) mid ni algo nr))

(defmacro (hash_3) (hash "prf" (ltk I R) mid ni nr))
(defmacro (hash_33) (hash "prf" (ltk R I) mid ni nr))

(defmacro (ilength) 2)
(defmacro (rlength) 3)

(defmacro (keymat1) (hash "prf" (ltk I R) lst algo ni nr))
(defmacro (keymat2) (hash "prf" (ltk R I) lst algo ni nr))


(defprotocol quick diffie-hellman
	     (defrole init
		       (vars (ni nr mid lst algo text) (I R name))
		       (trace
			 (send (cat I R mid (enc (hash_1i) lst ni (ltk I R))
				    (enc (hash_1ii) lst ni (ltk R I))))
			 (recv (cat R I mid (enc (hash_2i) algo nr (ltk I R))
				    (enc (hash_2ii) algo nr (ltk R I))))
			 (send (cat I R mid (enc (hash_3) (ltk I R))
				    (enc (hash_33) (ltk R I)))))
		       (uniq-orig ni mid))
	     (defrole resp
		      (vars (ni nr mid lst algo text) (I R name))
		      (trace
			(recv (cat I R mid (enc (hash_1r) lst ni (ltk I R))
				   (enc (hash_1rr) lst ni (ltk R I))))
			(send (cat R I mid (enc (hash_2r) algo nr (ltk I R))
				   (enc (hash_2rr) algo nr (ltk R I))))
			(recv (cat I R mid (enc (hash_3) (ltk I R))
				   (enc (hash_33) (ltk R I))))
                        (send "blab"))
			(uniq-orig nr))
             (defrole blab-skey
               (vars (key skey))
               (trace
                (recv "blab")
                (send key))
               (uniq-orig key)))

;;;
;;; Authentication skeletons
;;;

(defskeleton quick
	    (vars (ni nr ci  mid lst algo text) (I R name))
	    (defstrand init 3 (I I) (R R) (lst lst))
            (non-orig (ltk I R) (ltk R I)))

(defskeleton quick
	    (vars (ni nr ci mid lst algo text)  (I R name))
	    (defstrand resp 3 (I I) (R R) (lst lst) (algo algo) (nr nr))
            (non-orig (ltk I R) (ltk R  I)))

;;;
;;; Key Compromise Impersonation Skeletons
;;;

;;; This is not applicable to protocols relying only on shared secrets

;--------------------------------

(defgoal quick
  ; Aliveness - my peer has taken an action
  (forall ((iend node) (I R name) (ni nr mid lst algo text))
	  (implies
	   (and (p "init" (ilength) iend) 
		(p "init" "R" iend R) (p "init" "I" iend I)
		(p "init" "ni" iend ni)
		(p "init" "nr" iend nr)
		(p "init" "mid" iend mid)
		(p "init" "lst" iend lst)
		(p "init" "algo" iend algo)
		(non (ltk R I)) (non (ltk I R)))
	   (or (exists ((r1 node))
		       (and (p "resp" 0 r1) (p "resp" "R" r1 R)))
	       (exists ((i1 node))
		       (and (p "init" 0 i1) (p "init" "I" i1 R))))))
  ; Weak agreement - my peer and I agree on our names
  (forall ((iend node) (I R name) (ni nr mid lst algo text))
	  (implies
	   (and (p "init" (ilength) iend) 
		(p "init" "R" iend R) (p "init" "I" iend I)
		(p "init" "ni" iend ni)
		(p "init" "nr" iend nr)
		(p "init" "mid" iend mid)
		(p "init" "lst" iend lst)
		(p "init" "algo" iend algo)
		(non (ltk R I)) (non (ltk I R)))
	   (or (exists ((r1 node))
		       (and (p "resp" 0 r1) (p "resp" "R" r1 R) (p "resp" "I" r1 I)))
	       (exists ((i1 node))
		       (and (p "init" 0 i1) (p "init" "I" i1 R) (p "init" "R" i1 I))))))
  ; Minimal agreement - my peer and I agree on our names and roles, and on the DH key
  (forall ((iend node) (I R name) (ni nr mid lst algo text))
	  (implies
	   (and (p "init" (ilength) iend) 
		(p "init" "R" iend R) (p "init" "I" iend I)
		(p "init" "ni" iend ni)
		(p "init" "nr" iend nr)
		(p "init" "mid" iend mid)
		(p "init" "lst" iend lst)
		(p "init" "algo" iend algo)
		(non (ltk R I)) (non (ltk I R)))
	   (exists ((r1 node))
		   (and (p "resp" 0 r1) (p "resp" "R" r1 R) (p "resp" "I" r1 I)
			))))
  ; Full agreement - my peer and I agree on the full transcript
  (forall ((iend node) (I R name) (ni nr mid lst algo text))
	  (implies
	   (and (p "init" (ilength) iend) 
		(p "init" "R" iend R) (p "init" "I" iend I)
		(p "init" "ni" iend ni)
		(p "init" "nr" iend nr)
		(p "init" "mid" iend mid)
		(p "init" "lst" iend lst)
		(p "init" "algo" iend algo)
		(non (ltk R I)) (non (ltk I R)))
	   (exists ((r1 node))
		   (and (p "resp" 0 r1) (p "resp" "R" r1 R) (p "resp" "I" r1 I)
			(p "resp" "ni" r1 ni) (p "resp" "nr" r1 nr)
			(p "resp" "mid" r1 mid)
			(p "resp" "lst" r1 lst)
			(p "resp" "algo" r1 algo)))))
  )

(defgoal quick
  ; Aliveness - my peer has taken an action
  (forall ((rend node) (I R name) (ni nr mid lst algo text))
	  (implies
	   (and (p "resp" (rlength) rend) 
		(p "resp" "R" rend R) (p "resp" "I" rend I) (p "resp" "ni" rend ni)
		(p "resp" "nr" rend nr) (p "resp" "mid" rend mid)
		(p "resp" "lst" rend lst) (p "resp" "algo" rend algo)
		(non (ltk R I)) (non (ltk I R)))
	   (or (exists ((r1 node))
		       (and (p "resp" 0 r1) (p "resp" "R" r1 I)))
	       (exists ((i1 node))
		       (and (p "init" 0 i1) (p "init" "I" i1 I))))))
  ; Weak agreement - my peer and I agree on our names
  (forall ((rend node) (I R name) (ni nr mid lst algo text))
	  (implies
	   (and (p "resp" (rlength) rend) 
		(p "resp" "R" rend R) (p "resp" "I" rend I) (p "resp" "ni" rend ni)
		(p "resp" "nr" rend nr) (p "resp" "mid" rend mid)
		(p "resp" "lst" rend lst) (p "resp" "algo" rend algo)
		(non (ltk R I)) (non (ltk I R)))
	   (or (exists ((r1 node))
		       (and (p "resp" 0 r1) (p "resp" "R" r1 I) (p "resp" "I" r1 R)))
	       (exists ((i1 node))
		       (and (p "init" 0 i1) (p "init" "I" i1 I) (p "init" "R" i1 R))))))
  ; Minimal agreement - my peer and I agree on our names and roles, and on the DH key
  (forall ((rend node) (I R name) (ni nr mid lst algo text))
	  (implies
	   (and (p "resp" (rlength) rend) 
		(p "resp" "R" rend R) (p "resp" "I" rend I) (p "resp" "ni" rend ni)
		(p "resp" "nr" rend nr) (p "resp" "mid" rend mid)
		(p "resp" "lst" rend lst) (p "resp" "algo" rend algo)
		(non (ltk R I)) (non (ltk I R)))
	   (exists ((i1 node))
		   (and (p "init" 0 i1) (p "init" "R" i1 R) (p "init" "I" i1 I)))))
  ; Full agreement - my peer and I agree on the full transcript
  (forall ((rend node) (I R name) (ni nr mid lst algo text))
	  (implies
	   (and (p "resp" (rlength) rend) 
		(p "resp" "R" rend R) (p "resp" "I" rend I) (p "resp" "ni" rend ni)
		(p "resp" "nr" rend nr) (p "resp" "mid" rend mid)
		(p "resp" "lst" rend lst) (p "resp" "algo" rend algo)
		(non (ltk R I)) (non (ltk I R)))
	   (exists ((i1 node))
		   (and (p "init" 0 i1) (p "init" "R" i1 R) (p "init" "I" i1 I)
			(p "init" "ni" i1 ni) (p "init" "nr" i1 nr)
			(p "init" "mid" i1 mid)
			(p "init" "algo" i1 algo) (p "init" "lst" i1 lst)
			))))
  (comment "Responder point of view; agreement / weak agreement / aliveness goals")
  )

;;;
;;; Key secrecy skeletons (no privk leaks)
;;;

(defskeleton quick
	    (vars (ni nr ci cr lst algo text) (i expn) (Gr base) (I R name))
	    (defstrand init 3 (I I) (R R) (ni ni) (nr nr) (lst lst) (algo algo))
            (deflistener (keymat1))
	    (non-orig (ltk I R) (ltk R I)))

(defskeleton quick
	    (vars (ni nr ci cr lst algo text) (i expn) (Gr base) (I R name))
	    (defstrand init 3 (I I) (R R) (ni ni) (nr nr) (lst lst) (algo algo))
            (deflistener (keymat2))
	    (non-orig (ltk I R) (ltk R I)))

(defskeleton quick
	    (vars (ni nr ci cr lst algo text) (r expn) (Gi base) (I R name))
	    (defstrand resp 3 (I I) (R R) (ni ni) (nr nr) (lst lst) (algo algo))
            (deflistener (keymat1))
	    (non-orig (ltk I R) (ltk R I)))

(defskeleton quick
	    (vars (ni nr ci cr lst algo text) (r expn) (Gi base) (I R name))
	    (defstrand resp 3 (I I) (R R) (ni ni) (nr nr) (lst lst) (algo algo))
            (deflistener (keymat2))
	    (non-orig (ltk I R) (ltk R I)))



;;;
;;; Forward Secrecy skeletons (privk leaks)
;;;

;;; This protocol is not designed to have forward secrecy of any kind



