;created as an attempt to replicate results found by Cremers as expressed in Key Exchange in IPsec Revisited.
;(Or rather from input of ikev2-mac.cpp)
(herald "IKEv2 MAC" (algebra diffie-hellman) (bound 15))


;Macros to reflect what certian parties do not know
(defmacro (zi) (exp Gr i))
(defmacro (zr) (exp Gi r))

;Note: Instead of using the terms SPIi, SPIr 
;I have decided to stick with ci, cr as described in IKEv1
;to avoid confusion of terms

;Macros based on those for IKEV2 in common.h
(defmacro (SKi) (hash "KDF" ni nr (zi) ci cr))
(defmacro (SKr) (hash "KDF" ni nr (zr) ci cr))
;Macro for the weak ("perfect") forward secrecy
(defmacro (SKf) (hash "KDF" ni nr (exp (gen) (mul i r)) ci cr))

;Similar to the implementaion in IKEv1 I am using ltk to simulate what Cremers calls k(I,R) and k(R,I) in Scyther
;I have used the technique of sending duplicates to get the effects of bidirectional keys.
;You may want to change this later on
(defmacro (AUTHii) (hash "MAC" (ltk I R) ci "0" lst1 (exp (gen) i) ni nr (hash (SKi) I)))
(defmacro (AUTHiii) (hash "MAC" (ltk R I) ci "0" lst1 (exp (gen) i) ni nr (hash (SKi) I)))

(defmacro (AUTHir) (hash "MAC" (ltk I R) ci "0" lst1 Gi ni nr (hash (SKr) I)))
(defmacro (AUTHirr) (hash "MAC" (ltk R I) ci "0" lst1 Gi ni nr (hash (SKr) I)))

(defmacro (AUTHri) (hash "MAC" (ltk I R) ci cr algo1 Gr nr ni (hash (SKi) R)))
(defmacro (AUTHrii) (hash "MAC" (ltk R I) ci cr algo1 Gr nr ni (hash (SKi) R)))

(defmacro (AUTHrr) (hash "MAC" (ltk I R) ci cr algo1 (exp (gen) r) nr ni (hash (SKr) R)))
(defmacro (AUTHrrr) (hash "MAC" (ltk R I) ci cr algo1 (exp (gen) r) nr ni (hash (SKr) R)))

(defmacro (ilength) 4)
(defmacro (rlength) 3)

(defprotocol mac diffie-hellman
	     (defrole init
		      (vars (ci cr ni nr lst1 algo1 lst2 algo2 TSi TSr text) (Gr base) (i expn) (R I name))
		      (trace
			(send (cat I R ci "0" lst1 (exp (gen) i) ni))
			(recv (cat R I ci cr algo1 Gr nr))
			(send (cat I R ci cr (enc I R (AUTHii) lst2 TSi TSr (SKi)) 
				   (enc I R (AUTHiii) lst2 TSi TSr (SKi))))
			(recv (cat R I ci cr (enc R (AUTHri) algo2 TSi TSr (SKi)) 
				   (enc R (AUTHrii) algo2 TSi TSr (SKi))))
			(send "blab"))
		      (uniq-orig ni ci)
		      (uniq-gen i))
	     (defrole resp
		      (vars (ci cr ni nr lst1 algo1 lst2 algo2 TSi TSr text) (Gi base) (r expn) (R I name))
		      (trace
			(recv (cat I R ci "0" lst1 Gi ni))
			(send (cat R I ci cr algo1 (exp (gen) r) nr))
			(recv (cat I R ci cr (enc I R (AUTHir) lst2 TSi TSr (SKr))
				   (enc I R (AUTHirr) lst2 TSi TSr (SKr))))
			(send (cat R I ci cr (enc R (AUTHrr) algo2 TSi TSr (SKr))
				   (enc R (AUTHrrr) algo2 TSi TSr (SKr)))))
		      (uniq-orig nr cr)
		      (uniq-gen r))
	     (defrole blab-skey
               (vars (key skey))
               (trace
                (recv "blab")
                (send key))
               (uniq-orig key)))

;Standard skeletons:
(defgoal mac
  ; Aliveness - my peer has taken an action
  (forall ((iend node) (I R name) (ni nr ci cr lst1 lst2 algo1 algo2 TSi TSr
                  text) (i expn) (Gr base))
	  (implies
	   (and (p "init" (ilength) iend) 
	        (p "init" "R" iend R) (p "init" "I" iend I)
		(p "init" "ni" iend ni)	(p "init" "nr" iend nr)
		(p "init" "ci" iend ci)	(p "init" "cr" iend cr)
		(p "init" "lst1" iend lst1) (p "init" "lst2" iend lst2)
		(p "init" "i" iend i) (p "init" "Gr" iend Gr)
		(p "init" "algo1" iend algo1) (p "init" "algo2" iend algo2)
		(p "init" "TSi" iend TSi) (p "init" "TSr" iend TSr)
		(non (ltk R I)) (non (ltk I R)))
	   (or (exists ((r1 node))
		       (and (p "resp" 0 r1) (p "resp" "R" r1 R)))
	       (exists ((i1 node))
		       (and (p "init" 0 i1) (p "init" "I" i1 R))))))
  ; Weak agreement - my peer and I agree on our names
  (forall ((iend node) (I R name) (ni nr ci cr lst1 lst2 algo1 algo2 TSi TSr
                  text) (i expn) (Gr base))
	  (implies
	   (and (p "init" (ilength) iend) 
	        (p "init" "R" iend R) (p "init" "I" iend I)
		(p "init" "ni" iend ni)	(p "init" "nr" iend nr)
		(p "init" "ci" iend ci)	(p "init" "cr" iend cr)
		(p "init" "lst1" iend lst1) (p "init" "lst2" iend lst2)
		(p "init" "i" iend i) (p "init" "Gr" iend Gr)
		(p "init" "algo1" iend algo1) (p "init" "algo2" iend algo2)
		(p "init" "TSi" iend TSi) (p "init" "TSr" iend TSr)
		(non (ltk R I)) (non (ltk I R)))
	   (or (exists ((r1 node))
		       (and (p "resp" 0 r1) (p "resp" "R" r1 R) (p "resp" "I" r1 I)))
	       (exists ((i1 node))
		       (and (p "init" 0 i1) (p "init" "I" i1 R) (p "init" "R" i1 I))))))
  ; Minimal agreement - my peer and I agree on our names and roles, and on the DH key
  (forall ((iend node) (I R name) (ni nr ci cr lst1 lst2 algo1 algo2 TSi TSr
                  text) (i expn) (Gr base))
	  (implies
	   (and (p "init" (ilength) iend) 
	        (p "init" "R" iend R) (p "init" "I" iend I)
		(p "init" "ni" iend ni)	(p "init" "nr" iend nr)
		(p "init" "ci" iend ci)	(p "init" "cr" iend cr)
		(p "init" "lst1" iend lst1) (p "init" "lst2" iend lst2)
		(p "init" "i" iend i) (p "init" "Gr" iend Gr)
		(p "init" "algo1" iend algo1) (p "init" "algo2" iend algo2)
		(p "init" "TSi" iend TSi) (p "init" "TSr" iend TSr)
		(non (ltk R I)) (non (ltk I R)))
	   (exists ((r1 node) (Gi base) (r expn))
		   (and (p "resp" 0 r1) (p "resp" "R" r1 R) (p "resp" "I" r1 I)
			(p "resp" "Gi" r1 Gi) (p "resp" "r" r1 r)
			(= (exp Gi r) (exp Gr i))))))
  ; Full agreement - my peer and I agree on the full transcript
  (forall ((iend node) (I R name) (ni nr ci cr lst1 lst2 algo1 algo2 TSi TSr
                  text) (i expn) (Gr base))
	  (implies
	   (and (p "init" (ilength) iend) 
	        (p "init" "R" iend R) (p "init" "I" iend I)
		(p "init" "ni" iend ni)	(p "init" "nr" iend nr)
		(p "init" "ci" iend ci)	(p "init" "cr" iend cr)
		(p "init" "lst1" iend lst1) (p "init" "lst2" iend lst2)
		(p "init" "i" iend i) (p "init" "Gr" iend Gr)
		(p "init" "algo1" iend algo1) (p "init" "algo2" iend algo2)
		(p "init" "TSi" iend TSi) (p "init" "TSr" iend TSr)
		(non (ltk R I)) (non (ltk I R)))
	   (exists ((r1 node) (r expn))
		   (and (p "resp" 0 r1) (p "resp" "R" r1 R) (p "resp" "I" r1 I)
			(p "resp" "ni" r1 ni) (p "resp" "nr" r1 nr)
			(p "resp" "r" r1 r) (p "resp" "ci" r1 ci)
			(p "resp" "cr" r1 cr) (p "resp" "lst1" r1 lst1)
			(p "resp" "lst2" r1 lst2) (p "resp" "algo1" r1 algo1)
			(p "resp" "TSi" r1 TSi) (p "resp" "TSr" r1 TSr)
			(p "resp" "Gi" r1 (exp (gen) i))
			(p "resp" "algo2" r1 algo2)))))
  )

(defgoal mac
  ; Aliveness - my peer has taken an action
  (forall ((rend node) (I R name) (ni nr ci cr lst1 lst2 algo1 algo2 TSi TSr text)
	   (r expn) (Gi base))
	  (implies
	   (and (p "resp" (rlength) rend) 
		(p "resp" "R" rend R) (p "resp" "I" rend I) (p "resp" "ni" rend ni)
		(p "resp" "nr" rend nr) (p "resp" "ci" rend ci) (p "resp" "cr" rend cr)
		(p "resp" "lst1" rend lst1) (p "resp" "algo1" rend algo1) (p "resp" "r" rend r)
		(p "resp" "lst2" rend lst2) (p "resp" "algo2" rend algo2)
		(p "resp" "Gi" rend Gi) (p "resp" "TSi" rend TSi) (p "resp" "TSr" rend TSr)
		(non (ltk R I)) (non (ltk I R)))
	   (or (exists ((r1 node))
		       (and (p "resp" 0 r1) (p "resp" "R" r1 I)))
	       (exists ((i1 node))
		       (and (p "init" 0 i1) (p "init" "I" i1 I))))))
  ; Weak agreement - my peer and I agree on our names
  (forall ((rend node) (I R name) (ni nr ci cr lst1 lst2 algo1 algo2 TSi TSr text)
	   (r expn) (Gi base))
	  (implies
	   (and (p "resp" (rlength) rend) 
		(p "resp" "R" rend R) (p "resp" "I" rend I) (p "resp" "ni" rend ni)
		(p "resp" "nr" rend nr) (p "resp" "ci" rend ci) (p "resp" "cr" rend cr)
		(p "resp" "lst1" rend lst1) (p "resp" "algo1" rend algo1) (p "resp" "r" rend r)
		(p "resp" "lst2" rend lst2) (p "resp" "algo2" rend algo2)
		(p "resp" "Gi" rend Gi) (p "resp" "TSi" rend TSi) (p "resp" "TSr" rend TSr)
		(non (ltk R I)) (non (ltk I R)))
	   (or (exists ((r1 node))
		       (and (p "resp" 0 r1) (p "resp" "R" r1 I) (p "resp" "I" r1 R)))
	       (exists ((i1 node))
		       (and (p "init" 0 i1) (p "init" "I" i1 I) (p "init" "R" i1 R))))))
  ; Minimal agreement - my peer and I agree on our names and roles, and on the DH key
  (forall ((rend node) (I R name) (ni nr ci cr lst1 lst2 algo1 algo2 TSi TSr text)
	   (r expn) (Gi base))
	  (implies
	   (and (p "resp" (rlength) rend) 
		(p "resp" "R" rend R) (p "resp" "I" rend I) (p "resp" "ni" rend ni)
		(p "resp" "nr" rend nr) (p "resp" "ci" rend ci) (p "resp" "cr" rend cr)
		(p "resp" "lst1" rend lst1) (p "resp" "algo1" rend algo1) (p "resp" "r" rend r)
		(p "resp" "lst2" rend lst2) (p "resp" "algo2" rend algo2)
		(p "resp" "Gi" rend Gi) (p "resp" "TSi" rend TSi) (p "resp" "TSr" rend TSr)
		(non (ltk R I)) (non (ltk I R)))
	   (exists ((i1 node) (Gr base) (i expn))
		   (and (p "init" 0 i1) (p "init" "R" i1 R) (p "init" "I" i1 I)
		        (p "init" "Gr" i1 Gr) (p "init" "i" i1 i)
			(= (exp Gi r) (exp Gr i))
			))))
  ; Full agreement - my peer and I agree on the full transcript
  (forall ((rend node) (I R name) (ni nr ci cr lst1 lst2 algo1 algo2 TSi TSr text)
	   (r expn) (Gi base))
	  (implies
	   (and (p "resp" (rlength) rend) 
		(p "resp" "R" rend R) (p "resp" "I" rend I) (p "resp" "ni" rend ni)
		(p "resp" "nr" rend nr) (p "resp" "ci" rend ci) (p "resp" "cr" rend cr)
		(p "resp" "lst1" rend lst1) (p "resp" "algo1" rend algo1) (p "resp" "r" rend r)
		(p "resp" "lst2" rend lst2) (p "resp" "algo2" rend algo2)
		(p "resp" "Gi" rend Gi) (p "resp" "TSi" rend TSi) (p "resp" "TSr" rend TSr)
		(non (ltk R I)) (non (ltk I R)))
	   (exists ((i1 node) (i expn))
		   (and (p "init" 0 i1) (p "init" "R" i1 R) (p "init" "I" i1 I)
			(p "init" "ni" i1 ni) (p "init" "nr" i1 nr)
			(p "init" "ci" i1 ci) (p "init" "cr" i1 cr)
			(p "init" "algo1" i1 algo1) (p "init" "lst1" i1 lst1)
			(p "init" "algo2" i1 algo2) (p "init" "lst2" i1 lst2)
			(p "init" "TSr" i1 TSr) (p "init" "TSi" i1 TSi)
		        (p "init" "Gr" i1 (exp (gen) r)) (p "init" "i" i1 i)
			))))
  (comment "Responder point of view; agreement / weak agreement / aliveness goals")
  )

;Weakest: Keys are never compromised:
(defskeleton mac
	     (vars (ci cr ni nr lst1 algo1 lst2 algo2 TSi TSr text) (Gr base) (i expn) (R I name))
	     (defstrand init 4 (I I) (R R) (i i) (ci ci) (cr cr) (nr nr) (ni ni) (TSi TSi) (TSr TSr) (Gr Gr))
	     (deflistener (SKi))
	     (non-orig (ltk I R) (ltk R I)))
(defskeleton mac
	     (vars (ci cr ni nr lst1 algo1 lst2 algo2 TSi TSr text) (Gi base) (r expn) (R I name))
	     (defstrand resp 4 (I I) (R R) (r r) (ci ci) (cr cr) (ni ni) (nr nr) (Gi Gi) (TSi TSi) (TSr TSr))
	     (deflistener (SKr))
	     (non-orig (ltk I R) (ltk R I)))


;;; Weak Forward Secrecy
(defskeleton mac
  (vars (ci ni TSi TSr cr nr lst1 algo1 lst2 algo2 text) (R I name)
    (r i expn))
  (defstrand init 5 (ci ci) (cr cr) (ni ni) (nr nr) (lst1 lst1)
    (algo1 algo1) (lst2 lst2) (algo2 algo2) (TSi TSi) (TSr TSr) (R R)
    (I I) (Gr (exp (gen) r)) (i i))
  (defstrand resp 4 (ci ci) (cr cr) (ni ni) (nr nr) (lst1 lst1)
    (algo1 algo1) (lst2 lst2) (algo2 algo2) (TSi TSi) (TSr TSr) (R R)
    (I I) (Gi (exp (gen) i)) (r r))
  (defstrand blab-skey 2 (key (ltk R I)))
  (defstrand blab-skey 2 (key (ltk I R)))
  (deflistener (SKf))
  (precedes ((0 0) (1 0)) ((0 2) (1 2)) ((1 1) (0 1)) ((1 3) (0 3)) ((0 4) (2 0)) ((0 4) (3 0)))
  (ind-zero-in (r (exp (gen) i)))
  (pen-non-orig (ltk I R) (ltk R I))
  (uniq-gen r i)
  (uniq-orig ci ni cr nr))


;;; Perfect Forward Secrecy
(defskeleton mac
	     (vars (ci cr ni nr lst1 algo1 lst2 algo2 TSi TSr text) (Gr base) (i r expn) (R I name))
	     (defstrand init 5 (I I) (R R) (i i) (ci ci) (cr cr) (ni ni) (nr nr) (TSi TSi) (TSr TSr) 
			(lst1 lst1) (algo1 algo1) (lst2 lst2) (algo2 algo2) (Gr Gr))
	     (defstrand blab-skey 2 (key (ltk R I)))
             (defstrand blab-skey 2 (key (ltk I R)))
	     (deflistener (SKi))
	     (precedes ((0 4) (1 0)) ((0 4) (2 0)))
	     (pen-non-orig (ltk I R) (ltk R I)))

(defskeleton mac
	     (vars (ci cr ni nr lst1 algo1 lst2 algo2 TSi TSr text) (Gi base) (i r expn) (R I name))
	     (defstrand resp 4 (I I) (R R) (r r) (ci ci) (cr cr) (ni ni) (nr nr) (TSi TSi) (TSr TSr) 
			(lst1 lst1) (algo1 algo1) (lst2 lst2) (algo2 algo2) (Gi Gi))
	     (defstrand blab-skey 2 (key (ltk R I)))
             (defstrand blab-skey 2 (key (ltk I R)))
	     (deflistener (SKr))
	     (precedes ((0 3) (1 0)) ((0 3) (2 0)))
	     (pen-non-orig (ltk R I) (ltk I R)))
